@@ -64,6 +64,7 @@ import sys
 import traceback
 import os
 import functools
+import atexit
 from .abstractsession import AbstractSession
 from .event import Event
 from . import exception
@@ -75,12 +76,13 @@ from .topic import Topic
 from . import utils
 from .utils import get_handle
 from .compat import with_metaclass
+from .chandle import CHandle
 
-# pylint: disable=line-too-long,bad-continuation,useless-object-inheritance,too-many-lines
+# pylint: disable=line-too-long,too-many-lines
 # pylint: disable=protected-access,too-many-public-methods,bare-except,too-many-function-args
 
 @with_metaclass(utils.MetaClassForClassesWithEnums)
-class ServiceRegistrationOptions(object):
+class ServiceRegistrationOptions(CHandle):
     """Contains the options which can be specified when registering a service.
 
     To use non-default options to :meth:`~ProviderSession.registerService()`,
@@ -134,19 +136,11 @@ class ServiceRegistrationOptions(object):
 
     def __init__(self):
         """Create :class:`ServiceRegistrationOptions` with default options."""
-        self.__handle = internals.blpapi_ServiceRegistrationOptions_create()
-
-    def __del__(self):
-        try:
-            self.destroy()
-        except (NameError, AttributeError):
-            pass
-
-    def destroy(self):
-        """Destroy this :class:`ServiceRegistrationOptions`."""
-        if self.__handle:
-            internals.blpapi_ServiceRegistrationOptions_destroy(self.__handle)
-            self.__handle = None
+        selfhandle = internals.blpapi_ServiceRegistrationOptions_create()
+        super(ServiceRegistrationOptions, self).__init__(
+            selfhandle,
+            internals.blpapi_ServiceRegistrationOptions_destroy)
+        self.__handle = selfhandle
 
     def setGroupId(self, groupId):
         """Set the Group ID for the service to be registered.
@@ -160,18 +154,17 @@ class ServiceRegistrationOptions(object):
         are considered as Group Id.
         """
         internals.blpapi_ServiceRegistrationOptions_setGroupId(
-            self.__handle, groupId.encode('utf-8'))
-            # NOTE: we should convert groupId to bytes here because
-            # otherwise we'll get an error in SWIG wrapper.
+            self.__handle, groupId)
 
     def setServicePriority(self, priority):
-        """Set the priority with which a service will be registered.
+        """Set the priority with which a subscription service will be
+        registered.
 
         Args:
             priority (int): The service priority
 
-        Set the priority with which a service will be registered to the
-        specified ``priority``, where numerically greater values of
+        Set the priority with which a subscription service will be registered
+        to the specified ``priority``, where numerically greater values of
         ``priority`` indicate higher priorities. The behavior is undefined
         unless ``priority`` is non-negative. Note that while the values
         pre-defined in ``ServiceRegistrationOptions`` are suitable for use
@@ -179,6 +172,8 @@ class ServiceRegistrationOptions(object):
 
         By default, a service will be registered with priority
         :attr:`PRIORITY_HIGH`.
+
+        Note this has no effect for request-response or resolution services.
         """
         return internals.blpapi_ServiceRegistrationOptions_setServicePriority(
             self.__handle,
@@ -197,7 +192,8 @@ class ServiceRegistrationOptions(object):
     def getServicePriority(self):
         """
         Returns:
-            int: The value of the service priority in this instance.
+            int: The value of the priority for subscription services in this
+                 instance.
         """
         priority = \
             internals.blpapi_ServiceRegistrationOptions_getServicePriority(
@@ -254,10 +250,6 @@ class ServiceRegistrationOptions(object):
         return internals.blpapi_ServiceRegistrationOptions_getPartsToRegister(
                 self.__handle)
 
-    def _handle(self):
-        """Return the internal implementation."""
-        return self.__handle
-
 
 @with_metaclass(utils.MetaClassForClassesWithEnums)
 class ProviderSession(AbstractSession):
@@ -297,7 +289,7 @@ class ProviderSession(AbstractSession):
     __handlerProxy = None
 
     @staticmethod
-    def __dispatchEvent(sessionRef, eventHandle):
+    def __dispatchEvent(sessionRef, eventHandle): # pragma: no cover
         """Use sessions ref to dispatch an event"""
         try:
             session = sessionRef()
@@ -369,23 +361,16 @@ class ProviderSession(AbstractSession):
             get_handle(options),
             self.__handlerProxy,
             get_handle(eventDispatcher))
+
+        _destroy = internals.ProviderSession_destroyHelper
+        # note: AbstractSession destroy passes AbstractSession handle
+        _dtor = lambda hndl: _destroy(self.__handle, self.__handlerProxy)
+
+        atexit.register(self.stop) # we must stop session before shutdown
         AbstractSession.__init__(
             self,
-            internals.blpapi_ProviderSession_getAbstractSession(self.__handle))
-
-    def __del__(self):
-        try:
-            self.destroy()
-        except (NameError, AttributeError):
-            pass
-
-    def destroy(self):
-        """Destructor."""
-        if self.__handle:
-            internals.ProviderSession_destroyHelper(
-                self.__handle,
-                self.__handlerProxy)
-            self.__handle = None
+            internals.blpapi_ProviderSession_getAbstractSession(self.__handle),
+            _dtor)
 
     def start(self):
         """Start this :class:`Session` in synchronous mode.
@@ -459,6 +444,8 @@ class ProviderSession(AbstractSession):
         deadlock. Once a :class:`Session` has been stopped it can only be
         destroyed.
         """
+        if sys.version_info >= (3, 6):
+            atexit.unregister(self.stop)
         return internals.blpapi_ProviderSession_stop(self.__handle) == 0
 
     def stopAsync(self):
@@ -550,6 +537,9 @@ class ProviderSession(AbstractSession):
         an asynchronous :class:`ProviderSession` then this :class:`Event` may
         be processed by the registered :class:`Event` before
         :meth:`registerService()` has returned.
+
+        When ``identity`` is not provided, the session identity will be used if
+        it has been authorized.
         """
         if options is None:
             options = ServiceRegistrationOptions()
@@ -594,6 +584,9 @@ class ProviderSession(AbstractSession):
         :class:`~Event.SERVICE_STATUS` :class:`Event` which will be generated
         once the service has been successfully registered or registration has
         failed.
+
+        When ``identity`` is not provided, the session identity will be used if
+        it has been authorized.
         """
         if correlationId is None:
             correlationId = CorrelationId()
@@ -641,6 +634,9 @@ class ProviderSession(AbstractSession):
         asynchronous :class:`ProviderSession` then these :class:`Event`\ s may
         be processed by the registered ``eventHandler`` before
         :meth:`resolve()` has returned.
+
+        When ``identity`` is not provided, the session identity will be used if
+        it has been authorized.
         """
         resolutionList._addSession(self)
         _ExceptionUtil.raiseOnError(
@@ -676,6 +672,9 @@ class ProviderSession(AbstractSession):
         :attr:`AUTO_REGISTER_SERVICES` is specified :attr:`~Event.SERVICE_STATUS`
         events may also be generated before or after :meth:`resolveAsync()`
         returns.
+
+        When ``identity`` is not provided, the session identity will be used if
+        it has been authorized.
         """
         resolutionList._addSession(self)
         _ExceptionUtil.raiseOnError(
@@ -773,6 +772,9 @@ class ProviderSession(AbstractSession):
         asynchronous :class:`ProviderSession` then these :class:`Event`\ s may
         be processed by the registered ``eventHandler`` before
         :meth:`createTopics()` has returned.
+
+        When ``identity`` is not provided, the session identity will be used if
+        it has been authorized.
         """
         topicList._addSession(self)
         _ExceptionUtil.raiseOnError(
@@ -805,6 +807,9 @@ class ProviderSession(AbstractSession):
         asynchronous :class:`ProviderSession` then these :class:`Event`\ s may
         be processed by the registered ``eventHandler`` before
         :meth:`createTopics()` has returned.
+
+        When ``identity`` is not provided, the session identity will be used if
+        it has been authorized.
         """
         topicList._addSession(self)
         _ExceptionUtil.raiseOnError(
@@ -977,10 +982,6 @@ class ProviderSession(AbstractSession):
         finally:
             internals.delete_topicPtrArray(topicsCArray)
 
-    def _handle(self):
-        """Return the internal implementation."""
-        return self.__handle
-
 __copyright__ = """
 Copyright 2012. Bloomberg Finance L.P.
 
